package ghidraplugin;

import ghidra.app.plugin.PluginCategoryNames;
import ghidra.app.plugin.ProgramPlugin;
import ghidra.framework.plugintool.PluginInfo;
import ghidra.framework.plugintool.PluginTool;
import ghidra.framework.plugintool.util.PluginStatus;
import ghidra.program.model.listing.Program;
import ghidra.program.util.ProgramChangeRecord;
import ghidra.program.util.ProgramEvent;
import ghidra.util.Msg;
import ghidra.framework.model.*;
import ghidra.program.model.symbol.*;
import ghidra.program.model.listing.Data;
import ghidra.program.model.listing.Function;
import ghidra.program.util.*;
import ghidra.framework.cmd.Command;
import ghidra.framework.model.TransactionInfo;

@PluginInfo(
    status = PluginStatus.STABLE,
    packageName = "SymbolLogger",
    category = PluginCategoryNames.COMMON,
    shortDescription = "Logs symbol and function renames",
    description = "A plugin that monitors and logs all symbol and function rename events to the console, including undo/redo context",
    servicesRequired = {},
    servicesProvided = {}
)
public class SymbolRenameLoggerPlugin extends ProgramPlugin implements DomainObjectListener {

    public SymbolRenameLoggerPlugin(PluginTool tool) {
        super(tool);
    }

    @Override
    protected void init() {
        super.init();
        Msg.info(this, "SymbolRenameLoggerPlugin initialized");
    }

    @Override
    protected void programActivated(Program program) {
        super.programActivated(program);
        if (program != null) {
            program.addListener(this);
            Msg.info(this, "Started listening for rename events in program: " + program.getName());
        }
    }

    @Override
    protected void programDeactivated(Program program) {
        if (program != null) {
            program.removeListener(this);
            Msg.info(this, "Stopped listening for rename events in program: " + program.getName());
        }
        super.programDeactivated(program);
    }

    @Override
    public void domainObjectChanged(DomainObjectChangedEvent ev) {
        for (DomainObjectChangeRecord record : ev) {
            if (record instanceof ProgramChangeRecord) {
                ProgramChangeRecord progRecord = (ProgramChangeRecord) record;
                handleProgramChange(progRecord);
            }
        }
    }

    /**
     * Determines the context of the current operation (direct action, undo, or redo)
     */
    private String getOperationContext() {
        try {
            Program program = getCurrentProgram();
            if (program != null) {
                TransactionInfo currentTransaction = program.getCurrentTransactionInfo();
                if (currentTransaction != null) {
                    String description = currentTransaction.getDescription();
                    if (description != null) {
                        if (description.startsWith("Undo")) {
                            return "[UNDO] ";
                        } else if (description.startsWith("Redo")) {
                            return "[REDO] ";
                        }
                    }
                }
            }
        } catch (Exception e) {
            // Silently handle any exceptions in context detection
        }
        return ""; // Direct user action (no prefix)
    }

    /**
     * Enhanced logging method that includes undo/redo context
     */
    private void logWithContext(String message) {
        String context = getOperationContext();
        Msg.info(this, context + message);
    }

    private void handleProgramChange(ProgramChangeRecord record) {
        ProgramEvent eventType = (ProgramEvent)record.getEventType();
        
        Msg.info(this, "Event: " + record.toString());
        
        // Handle symbol rename events
        if (eventType == ProgramEvent.SYMBOL_RENAMED) {
            handleSymbolRenamed(record);
        }
        // Handle function rename events (functions are also symbols, but we can be more specific)
        else if (eventType == ProgramEvent.FUNCTION_CHANGED) {
            handleFunctionChanged(record);
        }
        // Handle symbol added events (in case we want to track new symbols)
        else if (eventType == ProgramEvent.SYMBOL_ADDED) {
            handleSymbolAdded(record);
        }
        // Handle symbol removed events
        else if (eventType == ProgramEvent.SYMBOL_REMOVED) {
            handleSymbolRemoved(record);
        }
        // Handle data type changes (includes global type definitions, structures, etc.)
        else if (eventType == ProgramEvent.DATA_TYPE_CHANGED) {
            handleDataTypeChanged(record);
        }
        // Handle data type added events
        else if (eventType == ProgramEvent.DATA_TYPE_ADDED) {
            handleDataTypeAdded(record);
        }
        // Handle data type removed events
        else if (eventType == ProgramEvent.DATA_TYPE_REMOVED) {
            handleDataTypeRemoved(record);
        }
        // Handle data type renamed events
        else if (eventType == ProgramEvent.DATA_TYPE_RENAMED) {
            handleDataTypeRenamed(record);
        }
        // Handle function signature changes (parameters, return types, etc.)
        else if (eventType == ProgramEvent.FUNCTION_CHANGED) {
            handleFunctionSignatureChanged(record);
        }
        // Handle transaction events for better undo/redo tracking
//        else if (eventType == ProgramEvent..TRANSACTION_STARTED) {
//            handleTransactionStarted(record);
//        }
//        else if (eventType == ProgramEvent.TRANSACTION_ENDED) {
//            handleTransactionEnded(record);
//        }
    }

    private void handleTransactionStarted(ProgramChangeRecord record) {
        String context = getOperationContext();
        if (!context.isEmpty()) {
            logWithContext("TRANSACTION STARTED");
        }
    }

    private void handleTransactionEnded(ProgramChangeRecord record) {
        String context = getOperationContext();
        if (!context.isEmpty()) {
            logWithContext("TRANSACTION ENDED");
        }
    }

    private void handleSymbolRenamed(ProgramChangeRecord record) {
        Object newValue = record.getNewValue();
        Object oldValue = record.getOldValue();
        
        String newName = (newValue != null) ? newValue.toString() : "null";
        String oldName = (oldValue != null) ? oldValue.toString() : "null";
        
        logWithContext(String.format("SYMBOL RENAMED: '%s' -> '%s' at address %s", 
            oldName, newName, record.getStart()));
        
        // Try to get additional information about the symbol
        try {
            Program program = getCurrentProgram();
            if (program != null) {
                SymbolTable symbolTable = program.getSymbolTable();
                Symbol[] symbols = symbolTable.getSymbols(record.getStart());
                
                for (Symbol symbol : symbols) {
                    if (symbol.getName().equals(newName)) {
                        String symbolType = getSymbolTypeDescription(symbol);
                        logWithContext(String.format("  Symbol type: %s, Namespace: %s", 
                            symbolType, symbol.getParentNamespace().getName()));
                        
                        // If it's a function, log additional function details
                        if (symbol.getSymbolType() == SymbolType.FUNCTION) {
                            Function function = (Function) symbol.getObject();
                            if (function != null) {
                                logWithContext(String.format("  Function signature: %s", 
                                    function.getPrototypeString(false, false)));
                            }
                        }
                        break;
                    }
                }
            }
        } catch (Exception e) {
            Msg.error(this, "Error getting symbol details: " + e.getMessage());
        }
    }

    private void handleFunctionChanged(ProgramChangeRecord record) {
        // This handles function-specific changes including renames and signature changes
        Object newValue = record.getNewValue();
        Object oldValue = record.getOldValue();
        
        if (newValue instanceof Function && oldValue instanceof Function) {
            Function newFunc = (Function) newValue;
            Function oldFunc = (Function) oldValue;
            
            // Check for name changes
            if (!newFunc.getName().equals(oldFunc.getName())) {
                logWithContext(String.format("FUNCTION RENAMED: '%s' -> '%s' at address %s", 
                    oldFunc.getName(), newFunc.getName(), newFunc.getEntryPoint()));
            }
            
            // Check for signature changes
            String oldSignature = oldFunc.getPrototypeString(false, false);
            String newSignature = newFunc.getPrototypeString(false, false);
            if (!oldSignature.equals(newSignature)) {
                logWithContext(String.format("FUNCTION SIGNATURE CHANGED at address %s:", newFunc.getEntryPoint()));
                logWithContext(String.format("  Old: %s", oldSignature));
                logWithContext(String.format("  New: %s", newSignature));
                
                // Log specific changes
                logFunctionSignatureDetails(oldFunc, newFunc);
            }
        } else if (newValue instanceof Function) {
            Function function = (Function) newValue;
            logWithContext(String.format("FUNCTION MODIFIED: %s at address %s", 
                function.getName(), function.getEntryPoint()));
            logWithContext(String.format("  Signature: %s", function.getPrototypeString(false, false)));
        }
    }

    private void handleFunctionSignatureChanged(ProgramChangeRecord record) {
        // Handle detailed function signature changes
        handleFunctionChanged(record); // Reuse the existing logic
    }

    private void handleDataTypeChanged(ProgramChangeRecord record) {
        Object newValue = record.getNewValue();
        Object oldValue = record.getOldValue();
        
        logWithContext(String.format("DATA TYPE CHANGED at address %s", record.getStart()));
        
        if (oldValue != null && newValue != null) {
            logWithContext(String.format("  Old type: %s", oldValue.toString()));
            logWithContext(String.format("  New type: %s", newValue.toString()));
        } else if (newValue != null) {
            logWithContext(String.format("  Type set to: %s", newValue.toString()));
        }
        
        // Try to get more context about the data type change
        try {
            Program program = getCurrentProgram();
            if (program != null && record.getStart() != null) {
                Data data = program.getListing().getDataAt(record.getStart());
                if (data != null) {
                    logWithContext(String.format("  Data label: %s", data.getLabel()));
                    logWithContext(String.format("  Data type: %s", data.getDataType().getName()));
                }
            }
        } catch (Exception e) {
            Msg.error(this, "Error getting data type details: " + e.getMessage());
        }
    }

    private void handleDataTypeAdded(ProgramChangeRecord record) {
        Object newValue = record.getNewValue();
        
        logWithContext(String.format("DATA TYPE ADDED at address %s", record.getStart()));
        if (newValue != null) {
            logWithContext(String.format("  Type: %s", newValue.toString()));
        }
    }

    private void handleDataTypeRemoved(ProgramChangeRecord record) {
        Object oldValue = record.getOldValue();
        
        logWithContext(String.format("DATA TYPE REMOVED at address %s", record.getStart()));
        if (oldValue != null) {
            logWithContext(String.format("  Removed type: %s", oldValue.toString()));
        }
    }

    private void handleDataTypeRenamed(ProgramChangeRecord record) {
        Object newValue = record.getNewValue();
        Object oldValue = record.getOldValue();
        
        String newName = (newValue != null) ? newValue.toString() : "null";
        String oldName = (oldValue != null) ? oldValue.toString() : "null";
        
        logWithContext(String.format("DATA TYPE RENAMED: '%s' -> '%s' at address %s", 
            oldName, newName, record.getStart()));
    }

    private void logFunctionSignatureDetails(Function oldFunc, Function newFunc) {
        try {
            // Check return type changes
            if (!oldFunc.getReturnType().equals(newFunc.getReturnType())) {
                logWithContext(String.format("    Return type changed: %s -> %s", 
                    oldFunc.getReturnType().getName(), newFunc.getReturnType().getName()));
            }
            
            // Check parameter count changes
            int oldParamCount = oldFunc.getParameterCount();
            int newParamCount = newFunc.getParameterCount();
            if (oldParamCount != newParamCount) {
                logWithContext(String.format("    Parameter count changed: %d -> %d", 
                    oldParamCount, newParamCount));
            }
            
            // Check individual parameter changes
            int maxParams = Math.max(oldParamCount, newParamCount);
            for (int i = 0; i < maxParams; i++) {
                if (i < oldParamCount && i < newParamCount) {
                    ghidra.program.model.listing.Parameter oldParam = oldFunc.getParameter(i);
                    ghidra.program.model.listing.Parameter newParam = newFunc.getParameter(i);
                    
                    if (!oldParam.getDataType().equals(newParam.getDataType())) {
                        logWithContext(String.format("    Parameter %d type changed: %s -> %s", 
                            i, oldParam.getDataType().getName(), newParam.getDataType().getName()));
                    }
                    
                    if (!oldParam.getName().equals(newParam.getName())) {
                        logWithContext(String.format("    Parameter %d name changed: %s -> %s", 
                            i, oldParam.getName(), newParam.getName()));
                    }
                } else if (i < newParamCount) {
                    ghidra.program.model.listing.Parameter newParam = newFunc.getParameter(i);
                    logWithContext(String.format("    Parameter %d added: %s %s", 
                        i, newParam.getDataType().getName(), newParam.getName()));
                } else {
                    ghidra.program.model.listing.Parameter oldParam = oldFunc.getParameter(i);
                    logWithContext(String.format("    Parameter %d removed: %s %s", 
                        i, oldParam.getDataType().getName(), oldParam.getName()));
                }
            }
            
            // Check calling convention changes
            if (!oldFunc.getCallingConventionName().equals(newFunc.getCallingConventionName())) {
                logWithContext(String.format("    Calling convention changed: %s -> %s", 
                    oldFunc.getCallingConventionName(), newFunc.getCallingConventionName()));
            }
            
        } catch (Exception e) {
            Msg.error(this, "Error logging function signature details: " + e.getMessage());
        }
    }

    private void handleSymbolAdded(ProgramChangeRecord record) {
        Object newValue = record.getNewValue();
        if (newValue instanceof Symbol) {
            Symbol symbol = (Symbol) newValue;
            String symbolType = getSymbolTypeDescription(symbol);
            logWithContext(String.format("SYMBOL ADDED: '%s' (%s) at address %s", 
                symbol.getName(), symbolType, symbol.getAddress()));
        }
    }

    private void handleSymbolRemoved(ProgramChangeRecord record) {
        Object oldValue = record.getOldValue();
        if (oldValue instanceof Symbol) {
            Symbol symbol = (Symbol) oldValue;
            String symbolType = getSymbolTypeDescription(symbol);
            logWithContext(String.format("SYMBOL REMOVED: '%s' (%s) at address %s", 
                symbol.getName(), symbolType, symbol.getAddress()));
        }
    }

    private String getSymbolTypeDescription(Symbol symbol) {
        SymbolType type = symbol.getSymbolType();
        
        if (type == SymbolType.FUNCTION) {
            return "Function";
        } else if (type == SymbolType.LABEL) {
            return "Label";
        } else if (type == SymbolType.CLASS) {
            return "Class";
        } else if (type == SymbolType.NAMESPACE) {
            return "Namespace";
        } else if (type == SymbolType.PARAMETER) {
            return "Parameter";
        } else if (type == SymbolType.LOCAL_VAR) {
            return "Local Variable";
        } else if (type == SymbolType.GLOBAL_VAR) {
            return "Global Variable";
        } else if (type == SymbolType.LIBRARY) {
            return "Library";
        } else {
            return type.toString();
        }
    }

    @Override
    protected void dispose() {
        Program program = getCurrentProgram();
        if (program != null) {
            program.removeListener(this);
        }
        super.dispose();
        Msg.info(this, "SymbolRenameLoggerPlugin disposed");
    }
} 